SystemOrganization addCategory: #'PostgresV3-CodeMirror'!
(PackageInfo named: 'PostgresV3-CodeMirror') postscript: '"Recompile all methods on the instance side of the subclasses of PG3SchemaMirror to ensure that the first literal of the methods contain the plpgsql source."
PG3SchemaMirror allSubclassesDo: #compileAll'!

SHParserST80 subclass: #PG3ShoutParser
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'PostgresV3-CodeMirror'!

!PG3ShoutParser methodsFor: 'parse' stamp: 'klub 7/4/2015 16:17'!
parseMethodTemporaries
	"No temporaries here; do nothing."! !

!PG3ShoutParser methodsFor: 'parse' stamp: 'klub 7/4/2015 16:18'!
parseStatementList
	"Set currentToken to nil, because we're ready."
	
	currentToken := nil! !

Compiler subclass: #PG3Compiler
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'PostgresV3-CodeMirror'!

!PG3Compiler methodsFor: 'as yet unclassified' stamp: 'kb 7/26/2010 17:50'!
parser

	^self parserClass new! !

!PG3Compiler methodsFor: 'as yet unclassified' stamp: 'kb 7/26/2010 17:35'!
parserClass

	^PG3Parser! !

Parser subclass: #PG3Parser
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'PostgresV3-CodeMirror'!

!PG3Parser methodsFor: 'as yet unclassified' stamp: 'klub 6/4/2014 16:56'!
checkSyntaxFor: functionPragma

	| functionMirror schemaMirror response |
	[
		schemaMirror := encoder classEncoding.
		functionMirror := PG3FunctionMirror 
			fromPragma: functionPragma 
			in: schemaMirror.
		functionMirror body: token.
		 schemaMirror executeTransaction: [ :connection |
			[ response := functionMirror saveToDatabase ] ensure: [
				(cue requestor notNil and: [ schemaMirror commitOnAccept ]) ifFalse: [ connection rollback ] ] ] ]
		on: Error
		do: [ :err |
			Transcript show: 'Couldn''t check syntax.'; cr.
			^self ].
	response ifEmpty: [ ^self ].
	(response last class == PG3ErrorResponse) ifTrue: [ 
		| query bodyIndex errorResponse errorPosition |
		errorResponse := response last.
		query := functionMirror printString.
		bodyIndex := (query indexOf: $$) + 2.
		errorResponse position
			ifNil: [ 
				errorPosition := errorResponse internalPosition
					ifNil: [ 0 ]
					ifNotNil: [ :internalPosition | internalPosition asInteger ] ]
			ifNotNil: [
				errorPosition := errorResponse position asInteger - bodyIndex.
				"Compensate for line ending differences"
				bodyIndex to: bodyIndex + errorPosition - 1 do: [ :index |
					(query at: index) = Character cr ifTrue: [
						errorPosition := errorPosition - 1 ] ] ].
		self 
			notify: errorResponse message 
			at: errorPosition + prevEnd + 2 ]
	! !

!PG3Parser methodsFor: 'as yet unclassified' stamp: 'klub 5/21/2014 16:47'!
functionPragma

	^properties pragmas 
		detect: [ :each | 
			PG3SchemaMirror acceptedPragmaSelectors includes: each keyword ]
		ifNone: nil! !

!PG3Parser methodsFor: 'as yet unclassified' stamp: 'klub 5/21/2014 16:39'!
messageNodeFor: functionPragma
	
	| refcursorFunction |
	refcursorFunction := functionPragma arguments third = 'refcursor'.
	^(MessageNode new
		receiver: (MessageNode new
			receiver: (encoder encodeVariable: 'self')
			selector: #class
			arguments: #()
			precedence: 1
			from: encoder)
		selector: (
			refcursorFunction
				ifTrue: [ #callRefcursorFunction:signature:withArguments:rowClass: ]
				ifFalse: [ #callNormalFunction:signature:withArguments: ])
		arguments: { 
			encoder encodeLiteral: functionPragma arguments first.
			encoder encodeLiteral: functionPragma arguments second.
			(self withArgumentArrayFor: functionPragma) 
				ifTrue: [ encoder tempNodes first ]
				ifFalse: [ BraceNode new elements: encoder tempNodes ] },
			(refcursorFunction
				ifTrue: [ { 
					(functionPragma arguments at: 5 ifAbsent: [ nil ]) 
						ifNil: [ encoder encodeLiteral: nil ]
						ifNotNil: [ :name | encoder encodeVariable: name ] } ]
				ifFalse: [ #() ])
		precedence: 3
		from: encoder) asReturnNode! !

!PG3Parser methodsFor: 'as yet unclassified' stamp: 'klub 9/12/2015 21:07'!
nodeForCode: aString

	^MessageNode new
		receiver: (encoder encodeLiteral: aString)
		selector: #class
		arguments: #()
		precedence: 1
		from: encoder! !

!PG3Parser methodsFor: 'as yet unclassified' stamp: 'kb 7/31/2010 16:35'!
statements: argNodes innerBlock: inner blockNode: theBlockNode

	| functionPragma |
	functionPragma := self functionPragma ifNil: [ 
		^super statements: argNodes innerBlock: inner blockNode: theBlockNode ].
	token := source position: prevEnd + 1; upToEnd.
	hereMark := hereEnd := source position + 1.
	hereType := #doIt.
	tokenType := #string.
	self checkSyntaxFor: functionPragma.
	theBlockNode
		arguments: argNodes
		statements: { 
			self nodeForCode: token.
			self messageNodeFor: functionPragma }
		returns: true
		from: encoder.
	parseNode := theBlockNode.
	^true! !

!PG3Parser methodsFor: 'as yet unclassified' stamp: 'klub 12/18/2010 15:06'!
withArgumentArrayFor: functionPragma

	^encoder tempNodes size = 1 and: [ 
		| size |
		size := functionPragma arguments second size.
		functionPragma arguments second second = 'refcursor' ifTrue: [ 
			" Refcursors as the first argument do not count into the arguments size "
			size := size - 2 ].
		size > 2 ]! !

Object subclass: #PG3Change
	instanceVariableNames: 'localFunction remoteFunction'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'PostgresV3-CodeMirror'!

!PG3Change methodsFor: 'as yet unclassified' stamp: 'kb 7/31/2010 16:22'!
deleteFromDatabase 

	remoteFunction deleteFromDatabase! !

!PG3Change methodsFor: 'as yet unclassified' stamp: 'kb 7/31/2010 16:57'!
diffText

	localFunction ifNil: [ ^remoteFunction printString asText ].
	^TextDiffBuilder
		buildDisplayPatchFrom: (remoteFunction ifNil: [ '' ] ifNotNil: [ remoteFunction printString ])
		to: localFunction printString! !

!PG3Change methodsFor: 'as yet unclassified' stamp: 'kb 7/29/2010 19:08'!
hasSameSource

	localFunction ifNil: [ ^false ].
	remoteFunction ifNil: [ ^false ].
	^localFunction printString = remoteFunction printString! !

!PG3Change methodsFor: 'as yet unclassified' stamp: 'kb 7/31/2010 15:37'!
isChanged

	^self isRemoteOnly not and: [ self isLocalOnly not ]! !

!PG3Change methodsFor: 'as yet unclassified' stamp: 'kb 7/31/2010 15:36'!
isLocalOnly

	^remoteFunction isNil! !

!PG3Change methodsFor: 'as yet unclassified' stamp: 'kb 7/31/2010 15:36'!
isRemoteOnly

	^localFunction isNil! !

!PG3Change methodsFor: 'as yet unclassified' stamp: 'kb 7/29/2010 18:41'!
localFunction

	^localFunction! !

!PG3Change methodsFor: 'as yet unclassified' stamp: 'kb 7/29/2010 18:41'!
localFunction: aPG3FunctionMirror

	localFunction := aPG3FunctionMirror! !

!PG3Change methodsFor: 'as yet unclassified' stamp: 'kb 7/31/2010 15:25'!
printOn: stream

	(localFunction ifNil: [ remoteFunction ]) printFunctionSignatureOn: stream.
	localFunction 
		ifNil: [ stream nextPutAll: ' (remote only)' ]
		ifNotNil: [ 
			remoteFunction 
				ifNil: [ stream nextPutAll: ' (local only)' ]
				ifNotNil: [ 
					self hasSameSource ifFalse: [
						stream nextPutAll: ' (changed)' ] ] ]! !

!PG3Change methodsFor: 'as yet unclassified' stamp: 'kb 7/29/2010 18:41'!
remoteFunction

	^remoteFunction! !

!PG3Change methodsFor: 'as yet unclassified' stamp: 'kb 7/29/2010 18:41'!
remoteFunction: aPG3FunctionMirror

	remoteFunction := aPG3FunctionMirror! !

!PG3Change methodsFor: 'as yet unclassified' stamp: 'kb 7/30/2010 16:53'!
saveIntoSchemaMirror

	remoteFunction saveIntoSchemaMirror! !

!PG3Change methodsFor: 'as yet unclassified' stamp: 'kb 7/30/2010 16:53'!
saveToDatabase

	localFunction saveToDatabase! !

Object subclass: #PG3FunctionMirror
	instanceVariableNames: 'schema name arguments returnType body language volatility schemaMirror'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'PostgresV3-CodeMirror'!

!PG3FunctionMirror class methodsFor: 'as yet unclassified' stamp: 'kb 7/29/2010 18:10'!
fromPragma: pragma in: schemaMirror

	^self new
		schemaMirror: schemaMirror;
		initializeFromPragma: pragma;
		yourself! !

!PG3FunctionMirror class methodsFor: 'as yet unclassified' stamp: 'kb 7/27/2010 12:35'!
fromRow: row in: schemaMirror

	^self new
		schemaMirror: schemaMirror;
		initializeFromRow: row;
		yourself! !

!PG3FunctionMirror methodsFor: 'accessing' stamp: 'kb 7/29/2010 21:45'!
argument: argumentName type: argumentType

	arguments 
		add: argumentName;
		add: (self typeNameFor: argumentType)! !

!PG3FunctionMirror methodsFor: 'as yet unclassified' stamp: 'kb 7/27/2010 13:32'!
arguments

	^arguments! !

!PG3FunctionMirror methodsFor: 'as yet unclassified' stamp: 'klub 1/30/2014 12:01'!
asSmalltalkArgumentName: aString

	^(String streamContents: [ :stream |
		(aString findTokens: '_') withIndexDo: [ :part :index |
			stream nextPutAll: (index = 1
				ifTrue: [ part ]
				ifFalse: [ part capitalized ] ) ] ]) asLegalSelector
			! !

!PG3FunctionMirror methodsFor: 'as yet unclassified' stamp: 'kb 7/27/2010 13:32'!
body

	^body! !

!PG3FunctionMirror methodsFor: 'as yet unclassified' stamp: 'kb 7/29/2010 18:18'!
body: aString

	body := aString withSqueakLineEndings! !

!PG3FunctionMirror methodsFor: 'as yet unclassified' stamp: 'kb 7/31/2010 16:15'!
compare: anotherFunction

	self name < anotherFunction name ifTrue: [ ^1 ].
	self name > anotherFunction name ifTrue: [ ^3 ].
	self arguments size < anotherFunction arguments size ifTrue: [ ^1 ].
	self arguments size > anotherFunction arguments size ifTrue: [ ^3 ].
	2 to: self arguments size by: 2 do: [ :argIndex |
		(self arguments at: argIndex) < (anotherFunction arguments at: argIndex)
			ifTrue: [ ^1].
		(self arguments at: argIndex) > (anotherFunction arguments at: argIndex)
			ifTrue: [ ^3] ].
	^2! !

!PG3FunctionMirror methodsFor: 'as yet unclassified' stamp: 'kb 8/4/2010 20:54'!
deleteFromDatabase

	schemaMirror executeQuery: (
		String streamContents: [ :stream |
			stream 
				nextPutAll: 'drop function ';
				nextPutAll: schemaMirror schema;
				nextPut: $..
			self printFunctionSignatureOn: stream ])! !

!PG3FunctionMirror methodsFor: 'as yet unclassified' stamp: 'kb 7/26/2010 17:05'!
immutable

	volatility := #immutable! !

!PG3FunctionMirror methodsFor: 'as yet unclassified' stamp: 'kb 7/27/2010 11:34'!
initialize

	super initialize.
	arguments := OrderedCollection new.
	language := 'plpgsql'.
	self volatile! !

!PG3FunctionMirror methodsFor: 'as yet unclassified' stamp: 'klub 9/13/2015 18:18'!
initializeFromPragma: pragma

	schema := schemaMirror schema.
	name := pragma argumentAt: 1.
	arguments := pragma argumentAt: 2.
	returnType := pragma argumentAt: 3.
	volatility := (pragma argumentAt: 4) asSymbol.
	pragma method ifNotNil: [ :method |
		self body: (method literalAt: 1) ]! !

!PG3FunctionMirror methodsFor: 'as yet unclassified' stamp: 'kb 7/29/2010 21:47'!
initializeFromRow: row

	#(schema name body) do: [ :each | 
		self perform: each asMutator with: (row atName: each) ].
	row argumentNames ifNotNil: [ :argumentNames |
		argumentNames with: row argumentTypes do: [ :arg :typ | 
			self argument: arg type: typ ] ].
	volatility := row volatility asSymbol.
	returnType := self typeNameFor: row returnType! !

!PG3FunctionMirror methodsFor: 'accessing' stamp: 'kb 7/27/2010 13:34'!
language

	^language! !

!PG3FunctionMirror methodsFor: 'as yet unclassified' stamp: 'kb 7/26/2010 17:04'!
language: aString

	language := aString! !

!PG3FunctionMirror methodsFor: 'as yet unclassified' stamp: 'kb 7/27/2010 13:33'!
name

	^name! !

!PG3FunctionMirror methodsFor: 'as yet unclassified' stamp: 'kb 7/26/2010 17:03'!
name: aString

	name := aString! !

!PG3FunctionMirror methodsFor: 'as yet unclassified' stamp: 'kb 7/27/2010 17:17'!
printBodyOn: stream

	stream nextPutAll: body! !

!PG3FunctionMirror methodsFor: 'as yet unclassified' stamp: 'kb 7/31/2010 15:26'!
printFunctionSignatureOn: stream

	stream
		nextPutAll: self name;
		nextPut: $(.
	2 to: arguments size by: 2 do: [ :index | 
		index = 2 ifFalse: [ stream nextPutAll: ', ' ].
		stream 
			nextPutAll: (arguments at: index) ].
	stream nextPut: $)! !

!PG3FunctionMirror methodsFor: 'as yet unclassified' stamp: 'kb 7/29/2010 21:46'!
printMethodPatternOn: stream

	| method |
	method := (schemaMirror methodForFunction: self) ifNil: [ 
		^self printSynthetizedMethodPatterOn: stream ].
	method selector isUnary ifTrue: [ 
		^stream nextPutAll: method selector ].
	method selector isInfix ifTrue: [ self error: 'Binary selectors not supported.' ].
	method selector keywords 
		with: (method methodNode tempNames first: method selector keywords size)
		do: [ :keyword :tempName | 
			stream 
				nextPutAll: keyword;
				space;
				nextPutAll: tempName;
				space ]! !

!PG3FunctionMirror methodsFor: 'as yet unclassified' stamp: 'kb 7/27/2010 17:09'!
printOn: stream

	stream 
		nextPutAll: 'create or replace function ';
		nextPutAll: schema;
		nextPut: $.;
		nextPutAll: name;
		nextPut: $(.
	arguments doWithIndex: [ :each :index | 
		index even
			ifTrue: [ stream nextPutAll: each ]  
			ifFalse: [
				index = 1 ifFalse: [ stream nextPutAll: ', ' ].
				stream 
					nextPutAll: each;
					space ] ].
	stream 
		nextPut: $);
		nextPutAll: ' returns ';
		nextPutAll: returnType;
		nextPutAll: ' as $$';
		nextPutAll: body;
		nextPutAll: '$$ ';
		nextPutAll: 'language ';
		print: language;
		space;
		nextPutAll: volatility asUppercase;
		nextPut: $;! !

!PG3FunctionMirror methodsFor: 'as yet unclassified' stamp: 'klub 2/11/2016 16:07'!
printPragmaOn: stream

	stream nextPut: $<.
	stream 
		nextPutAll: 'pg3Function: ' ;
		print: name;
		cr; tab; tab;
		nextPutAll: 'arguments: ';
		print: (arguments asArray withIndexCollect: [ :name :index |
			index even
				ifFalse: [ name ]
				ifTrue: [ name asSymbol ] ]);
		cr; tab; tab;
		nextPutAll: 'returns: ';
		print: returnType asSymbol;
		cr; tab; tab;
		nextPutAll: 'volatility: ';
		print: volatility;
		nextPut: $>;
		cr! !

!PG3FunctionMirror methodsFor: 'as yet unclassified' stamp: 'klub 1/30/2014 12:03'!
printSynthetizedMethodPatterOn: stream

	| nameParts firstRealArgumentIndex |
	nameParts := (name findTokens: '_').
	stream nextPutAll: nameParts first asLegalSelector.
	nameParts allButFirstDo: [ :each | 
		stream nextPutAll: each asLegalSelector capitalized ].
	firstRealArgumentIndex := (arguments size > 0 and: [ arguments second = 'refcursor' ])
		ifTrue: [ 3 ]
		ifFalse: [ 1 ].
	firstRealArgumentIndex to: arguments size by: 2 do: [ :index |
		| argumentName |
		argumentName := self asSmalltalkArgumentName: (arguments at: index).
		index = firstRealArgumentIndex 
			ifFalse: [ stream space; nextPutAll: argumentName ].
		stream nextPutAll: ': '; nextPutAll: argumentName ]! !

!PG3FunctionMirror methodsFor: 'accessing' stamp: 'kb 7/27/2010 13:34'!
returnType

	^returnType! !

!PG3FunctionMirror methodsFor: 'as yet unclassified' stamp: 'klub 2/11/2016 15:52'!
returnType: aString

	returnType := self typeNameFor: aString! !

!PG3FunctionMirror methodsFor: 'as yet unclassified' stamp: 'kb 7/29/2010 17:31'!
saveIntoSchemaMirror

	schemaMirror
		compile: (
			String streamContents: [ :stream |
				self printMethodPatternOn: stream.
				stream cr; cr; tab.
				self
					printPragmaOn: stream;
					printBodyOn: stream ])
		classified: 'functions'! !

!PG3FunctionMirror methodsFor: 'as yet unclassified' stamp: 'klub 1/7/2011 20:01'!
saveToDatabase

	^schemaMirror executeQuery: self printString withInternetLineEndings
	! !

!PG3FunctionMirror methodsFor: 'accessing' stamp: 'kb 7/27/2010 13:34'!
schema

	^schema! !

!PG3FunctionMirror methodsFor: 'as yet unclassified' stamp: 'kb 7/26/2010 17:03'!
schema: aString

	schema := aString! !

!PG3FunctionMirror methodsFor: 'accessing' stamp: 'kb 7/27/2010 13:34'!
schemaMirror

	^schemaMirror! !

!PG3FunctionMirror methodsFor: 'as yet unclassified' stamp: 'kb 7/27/2010 12:35'!
schemaMirror: aPG3SchemaMirrorClass

	schemaMirror := aPG3SchemaMirrorClass! !

!PG3FunctionMirror methodsFor: 'as yet unclassified' stamp: 'kb 7/26/2010 17:05'!
stable

	volatility := #stable! !

!PG3FunctionMirror methodsFor: 'as yet unclassified' stamp: 'klub 2/11/2016 15:57'!
typeNameFor: aString

	^#(
		('int4' 'integer')
		('_int4' 'integer[]')
		('int2' 'smallint')
		('_int2' 'smallint[]')
		('int8' 'bigint')
		('_int8' 'bigint[]')
		('bool' 'boolean'))
			detect: [ :each | each first = aString ] 
			ifFound: [ :each | each second ]
			ifNone: [ 
				aString first = $_ 
					ifFalse: [ aString ]
					ifTrue: [ aString allButFirst, '[]' ] ]! !

!PG3FunctionMirror methodsFor: 'as yet unclassified' stamp: 'kb 7/26/2010 17:05'!
volatile

	volatility := #volatile! !

!PG3FunctionMirror methodsFor: 'as yet unclassified' stamp: 'kb 7/27/2010 13:33'!
volatility

	^volatility! !

Object subclass: #PG3SchemaChanges
	instanceVariableNames: 'schemaMirror selected changes'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'PostgresV3-CodeMirror'!

!PG3SchemaChanges methodsFor: 'toolbuilder' stamp: 'kb 8/4/2010 17:27'!
buildWith: builder

	|  windowSpec listSpec panelSpec buttonSpec textSpec |
	windowSpec := builder pluggableWindowSpec new.
	windowSpec label: 'Differences of schema ', schemaMirror schema, ' and ', schemaMirror className.
	windowSpec model: self.
	windowSpec children: OrderedCollection new.
	
	selected := changes at: 1 ifAbsent: nil.	
	listSpec := builder pluggableListSpec new.
	listSpec 
		model: self;
		list: #changes; 
		getSelected: #selected; 
		setSelected: #selected:; 
		menu: #changeMenu:shifted:;
		frame: (0@0 corner: 1@0.3).
	windowSpec children add: listSpec.

	panelSpec := builder pluggablePanelSpec new.
	panelSpec children: OrderedCollection new.
	panelSpec frame: (0@0.3 corner: 1@0.38).

	buttonSpec := builder pluggableButtonSpec new.
	buttonSpec 
			model: self;
			label: 'Download All'; 
			action: #downloadAll;
			frame: (0@0 corner: 0.32@1).
	panelSpec children addLast: buttonSpec.

	buttonSpec := builder pluggableButtonSpec new.
	buttonSpec 
			model: self;
			label: 'Refresh'; 
			action: #refresh;
			frame: (0.33@0 corner: 0.65@1).
	panelSpec children addLast: buttonSpec.

	buttonSpec := builder pluggableButtonSpec new.
	buttonSpec 
			model: self;
			label: 'Upload All'; 
			action: #uploadAll;
			frame: (0.66@0 corner: 1@1).
	panelSpec children addLast: buttonSpec.
		
	windowSpec children add: panelSpec.

	textSpec := builder pluggableTextSpec new.
	textSpec 
		model: self;
		getText: #diffText; 
		frame: (0@0.38 corner: 1@1).
	windowSpec children add: textSpec.
	
	^builder build: windowSpec! !

!PG3SchemaChanges methodsFor: 'toolbuilder' stamp: 'kb 7/30/2010 17:25'!
buildWith: builder in: window rect: rect
	| listSpec textSpec |
	selected := changes first.	
	listSpec := builder pluggableListSpec new.
	listSpec 
		model: self;
		list: #changes; 
		getSelected: #selected; 
		setSelected: #selected:; 
		"menu: #changeSetMenu:shifted:; 
		keyPress: #changeSetListKey:from:;
		autoDeselect: false;"
		frame: (((0@0 extent: 1@0.5)
			scaleBy: rect extent) translateBy: rect origin).
	window children add: listSpec.

	textSpec := builder pluggableTextSpec new.
	textSpec 
		model: self;
		getText: #diffText; 
		"setText: #contents:notifying:; 
		selection: #contentsSelection; 
		menu: #codePaneMenu:shifted:;"
		frame: (((0@0.5 corner: 1@1) scaleBy: rect extent) translateBy: rect origin).
	window children add: textSpec.
	^window! !

!PG3SchemaChanges methodsFor: 'as yet unclassified' stamp: 'kb 7/31/2010 16:22'!
changeMenu: aMenu shifted: isShifted

	selected ifNil: [ ^nil ].
	selected isLocalOnly ifFalse: [
		aMenu add: 'download' action: #downloadSelected ].
	selected isRemoteOnly ifFalse: [
		aMenu add: 'upload' action: #uploadSelected ].
	selected isLocalOnly ifFalse: [
		aMenu add: 'delete remote' action: #deleteSelectedFromDatabase ].
	^aMenu! !

!PG3SchemaChanges methodsFor: 'toolbuilder' stamp: 'kb 7/30/2010 17:09'!
changes

	^changes! !

!PG3SchemaChanges methodsFor: 'as yet unclassified' stamp: 'kb 7/31/2010 16:23'!
deleteSelectedFromDatabase 

	selected deleteFromDatabase.
	self refresh! !

!PG3SchemaChanges methodsFor: 'toolbuilder' stamp: 'kb 7/31/2010 15:31'!
diffText

	^(selected ifNil: [ ^'' ]) diffText! !

!PG3SchemaChanges methodsFor: 'as yet unclassified' stamp: 'kb 7/31/2010 15:43'!
downloadAll

	changes do: [ :each | 
		each isLocalOnly ifFalse: [ 
			each saveIntoSchemaMirror ] ].
	self refresh! !

!PG3SchemaChanges methodsFor: 'as yet unclassified' stamp: 'kb 7/31/2010 15:38'!
downloadSelected

	selected saveIntoSchemaMirror.
	self refresh! !

!PG3SchemaChanges methodsFor: 'as yet unclassified' stamp: 'kb 7/31/2010 15:10'!
findChanges

	| matches |
	matches := self findMatches.
	^matches reject: #hasSameSource! !

!PG3SchemaChanges methodsFor: 'as yet unclassified' stamp: 'klub 5/21/2014 16:48'!
findMatches

	| localFunctions remoteFunctions allFunctions |
	localFunctions := schemaMirror fetchFunctionsFromCode.
	remoteFunctions := schemaMirror fetchFunctionsFromDatabase.
	allFunctions := OrderedCollection new.
	allFunctions addAll: (localFunctions collect: [ :each | each -> #local ]).
	allFunctions addAll: (remoteFunctions collect: [ :each | each -> #remote ]).
	allFunctions sort: [ :a :b | 
		| result |
		result := a key compare: b key.
		result = 1 or: [ 
			result = 2 and: [ 
				a value <= b value ] ] ].
	^Array streamContents: [ :stream |
		| index change |
		index := 2.
		[ index <= allFunctions size ] whileTrue: [
			| currentFunction previousFunction |
			previousFunction := (allFunctions at: index - 1) key.
			currentFunction := (allFunctions at: index) key.
			(previousFunction name = currentFunction name and: [
				previousFunction arguments size = currentFunction arguments size and: [ 
					(2 to: currentFunction arguments size by: 2) allSatisfy: [ :each |
						(currentFunction arguments at: each) = (previousFunction arguments at: each) ] ] ])
				ifFalse: [
					change := PG3Change new.
					(allFunctions at: index - 1) value == #local
						ifTrue: [ change localFunction: previousFunction ]
						ifFalse: [ change remoteFunction: previousFunction ].
					index := index + 1 ]
				ifTrue: [ 
					change := PG3Change new
						localFunction: previousFunction;
						remoteFunction: currentFunction;
						yourself.
					index := index + 2 ].
			stream nextPut: change ].
		index = (allFunctions size + 1) ifTrue: [
			change := PG3Change new.
					(allFunctions at: index - 1) value == #local
						ifTrue: [ change localFunction: (allFunctions at: index - 1) key ]
						ifFalse: [ change remoteFunction: (allFunctions at: index - 1) key ].
			stream nextPut: change ] ]! !

!PG3SchemaChanges methodsFor: 'toolbuilder' stamp: 'kb 7/31/2010 15:16'!
refresh

	| index |
	index := selected ifNotNil: [ changes indexOf: selected ] ifNil: [ 1 ].
	changes := self findChanges.
	self selected: (changes at: index ifAbsent: nil).
	self changed: #changes! !

!PG3SchemaChanges methodsFor: 'as yet unclassified' stamp: 'kb 7/31/2010 15:09'!
schemaMirror: aPG3SchemaMirror

	schemaMirror := aPG3SchemaMirror.
	changes := self findChanges! !

!PG3SchemaChanges methodsFor: 'toolbuilder' stamp: 'kb 7/30/2010 17:25'!
selected

	^selected! !

!PG3SchemaChanges methodsFor: 'toolbuilder' stamp: 'kb 7/30/2010 17:28'!
selected: aPG3Change

	selected := aPG3Change.
	self changed: #diffText! !

!PG3SchemaChanges methodsFor: 'as yet unclassified' stamp: 'kb 7/31/2010 15:43'!
uploadAll

	changes do: [ :each | 
		each isRemoteOnly ifFalse: [ 
			each saveToDatabase ] ].
	self refresh! !

!PG3SchemaChanges methodsFor: 'as yet unclassified' stamp: 'kb 7/31/2010 15:38'!
uploadSelected

	selected saveToDatabase.
	self refresh! !

Object subclass: #PG3SchemaMirror
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'PostgresV3-CodeMirror'!

PG3SchemaMirror class
	instanceVariableNames: 'instance commitOnAccept'!

PG3SchemaMirror class
	instanceVariableNames: 'instance commitOnAccept'!

!PG3SchemaMirror class methodsFor: 'compiling' stamp: 'klub 5/21/2014 16:47'!
acceptedPragmaSelectors

	^#(
		pg3Function:arguments:returns:volatility:rowClass:
		pg3Function:arguments:returns:volatility:
		pg3Function:arguments:returns:	)! !

!PG3SchemaMirror class methodsFor: 'synchronization' stamp: 'kb 7/30/2010 17:09'!
browseChanges

	ToolBuilder open: self changes! !

!PG3SchemaMirror class methodsFor: 'query execution' stamp: 'kb 8/4/2010 23:58'!
callNormalFunction: functionName signature: signature withArguments: arguments

	| result rs query |
	query := self createNormalQuery: functionName signature: signature withArguments: arguments.
	rs := self executeQuery: query.
	(rs last class = PG3ErrorResponse) ifTrue: [ 
		Error signal: rs last message ].
	result := rs last rows at: 1 ifAbsent: nil.
	result ifNotNil: [ result := result at: 1 ].
	^result! !

!PG3SchemaMirror class methodsFor: 'query execution' stamp: 'kb 8/4/2010 20:47'!
callRefcursorFunction: functionName signature: signature withArguments: arguments

	| rows rs query |
	query := self createRefcursorQuery: functionName signature: signature withArguments: arguments.
	rs := self executeQuery: query.
	(rs last class = PG3ErrorResponse) ifTrue: [ 
		Error signal: rs last message ].
	rows := (rs at: 2) rowsWithColumnNameDictionary.
	^rows! !

!PG3SchemaMirror class methodsFor: 'query execution' stamp: 'klub 5/21/2014 16:12'!
callRefcursorFunction: functionName signature: signature withArguments: arguments rowClass: aClass

	| rows rs query |
	query := self createRefcursorQuery: functionName signature: signature withArguments: arguments.
	
	rs := self executeQuery: query rowClass: aClass.
	(rs last class = PG3ErrorResponse) ifTrue: [ 
		Error signal: rs last message ].
	rows := (rs at: 2) rowsWithColumnNameDictionary.
	^rows! !

!PG3SchemaMirror class methodsFor: 'synchronization' stamp: 'kb 7/29/2010 18:33'!
changes

	^PG3SchemaChanges new
		schemaMirror: self;
		yourself! !

!PG3SchemaMirror class methodsFor: 'synchronization' stamp: 'kb 8/30/2010 17:54'!
commitOnAccept

	^commitOnAccept ifNil: [ true ]! !

!PG3SchemaMirror class methodsFor: 'synchronization' stamp: 'kb 8/2/2010 21:35'!
commitOnAccept: aBoolean
	" If true then save the edited function when accepted in the browser. "

	commitOnAccept := aBoolean! !

!PG3SchemaMirror class methodsFor: 'compiling' stamp: 'kb 7/26/2010 17:55'!
compilerClass

	^PG3Compiler! !

!PG3SchemaMirror class methodsFor: 'query building' stamp: 'klub 6/15/2014 23:51'!
createNormalQuery: functionName signature: signature withArguments: arguments

	signature size / 2 = arguments size ifFalse: [ 
		self error: 'Unexpected number of arguments.' ].
	^String streamContents: [ :stream |
		| argumentStream first |
		stream
			nextPutAll: 'select ';
			nextPutAll: self schema;
			nextPut: $.;
			nextPutAll: functionName;
			nextPutAll: '('.
		argumentStream := arguments readStream.
		first := true.
		signature pairsDo: [ :name :type |
			first 
				ifTrue: [ first := false ]
				ifFalse: [ stream nextPutAll: ', ' ].
			argumentStream next pg3PrintAsLiteralOn: stream.
			stream
				nextPutAll: '::';
				nextPutAll: type ].
		stream nextPutAll: ');' ]! !

!PG3SchemaMirror class methodsFor: 'query building' stamp: 'klub 9/12/2015 21:15'!
createRefcursorQuery: functionName signature: signature withArguments: arguments

	signature size - 2 / 2 = arguments size ifFalse: [ 
		self error: 'Unexpected number of arguments.' ].
	^String streamContents: [ :stream |
		| argumentStream first |
		stream
			nextPutAll: 'select ';
			nextPutAll: self schema;
			nextPut: $.;
			nextPutAll: functionName;
			nextPutAll: '('.
		argumentStream := arguments readStream.
		first := true.
		signature pairsDo: [ :name :type |
			first 
				ifTrue: [ first := false ]
				ifFalse: [ stream nextPutAll: ', ' ].
			type = 'refcursor'
				ifTrue: [ 
					self cursorName pg3PrintAsLiteralOn: stream.
					stream nextPutAll: '::refcursor' ]
				ifFalse: [
					argumentStream next pg3PrintAsLiteralOn: stream.
					stream
						nextPutAll: '::';
						nextPutAll: type ] ].
		stream nextPutAll: ');fetch all from '.
		self cursorName pg3PrintAsSqlIdentifierOn: stream.
		stream nextPutAll: ';close '.
		self cursorName pg3PrintAsSqlIdentifierOn: stream ]! !

!PG3SchemaMirror class methodsFor: 'parameters' stamp: 'kb 8/4/2010 20:16'!
cursorName
		
	^'squeakcursor'! !

!PG3SchemaMirror class methodsFor: 'as yet unclassified' stamp: 'kb 8/4/2010 20:24'!
default

	^instance ifNil: [ instance := self new ]! !

!PG3SchemaMirror class methodsFor: 'as yet unclassified' stamp: 'klub 5/21/2014 16:17'!
executeQuery: aBlock

	^self executeQuery: aBlock rowClass: nil! !

!PG3SchemaMirror class methodsFor: 'as yet unclassified' stamp: 'klub 5/21/2014 16:12'!
executeQuery: aBlock rowClass: aClass

	^self pool executeQuery: aBlock rowClass: aClass! !

!PG3SchemaMirror class methodsFor: 'as yet unclassified' stamp: 'kb 8/4/2010 20:48'!
executeTransaction: aBlock

	^self pool executeTransaction: aBlock! !

!PG3SchemaMirror class methodsFor: 'synchronization' stamp: 'klub 5/21/2014 16:47'!
fetchFunctionsFromCode

	^Array streamContents: [ :stream |
		self acceptedPragmaSelectors do: [ :keyword |
			(Pragma allNamed: keyword in: self) do: [ :pragma | 
				stream nextPut: (PG3FunctionMirror fromPragma: pragma in: self) ] ] ]! !

!PG3SchemaMirror class methodsFor: 'synchronization' stamp: 'klub 1/20/2016 22:14'!
fetchFunctionsFromDatabase

	| rs |
	rs := self pool withConnectionDo: [ :connection | 
		connection execute: '
select 
	pgn.nspname as "schema", 
	pgp.proname as "name",
	pgp.prosrc as "body",
	pgp.proargnames as "argumentNames",
	array(select
		case 
			when n.oid < 100 then t.typname
			else n.nspname || ''.'' || t.typname
		end
	 from 
		generate_series(array_lower(pgp.proargtypes, 1), array_upper(pgp.proargtypes, 1)) as s(a)
		join pg_type t on t.oid = pgp.proargtypes[s.a]
		join pg_namespace n on n.oid = t.typnamespace) as "argumentTypes",
	case 
		when pgretn.oid < 100 then pgrett.typname 
		else pgretn.nspname || ''.'' || pgrett.typname 
	end as "returnType",
	case pgp.provolatile 
		when ''v'' then ''volatile''
		when ''s'' then ''stable''
		when ''i'' then ''immutable''
	end as "volatility"
from 
	pg_proc pgp
	join pg_namespace pgn on pgn.oid = pgp.pronamespace
	join pg_type pgrett on pgrett.oid = pgp.prorettype
	join pg_namespace pgretn on pgretn.oid = pgrett.typnamespace
where 
	not pgp.proretset and
	pgn.nspname = ', self schema asPG3SqlLiteral ].
	^rs first rowsWithColumnNameDictionary collect: [ :each | 
		PG3FunctionMirror fromRow: each in: self ]! !

!PG3SchemaMirror class methodsFor: 'synchronization' stamp: 'klub 5/21/2014 16:47'!
methodForFunction: function

	self acceptedPragmaSelectors do: [ :keyword |
		((Pragma allNamed: keyword in: self) 
			detect: [ :each |self pragma: each matchesFunction: function ]
			ifNone: nil) ifNotNil: [ :pragma |
				^pragma method ] ].
	^nil! !

!PG3SchemaMirror class methodsFor: 'compiling' stamp: 'kb 7/29/2010 21:37'!
parserClass

	^PG3Parser! !

!PG3SchemaMirror class methodsFor: 'parameters' stamp: 'klub 7/7/2011 16:53'!
pool

	self subclassResponsibility! !

!PG3SchemaMirror class methodsFor: 'synchronization' stamp: 'kb 7/29/2010 19:18'!
pragma: pragma matchesFunction: function

	| arguments |
	(pragma argumentAt: 1) = function name ifFalse: [ ^false ].
	(arguments := (pragma argumentAt: 2)) size = function arguments size ifFalse: [ ^false ].
	2 to: arguments size by: 2 do: [ :i | 
		(arguments at: i) = (function arguments at: i) ifFalse: [ ^false ] ].
	^true! !

!PG3SchemaMirror class methodsFor: 'parameters' stamp: 'kb 7/26/2010 15:02'!
schema

	self subclassResponsibility! !

!PG3SchemaMirror class methodsFor: 'accessing' stamp: 'kb 7/29/2010 17:51'!
shoutParserClass

	^PG3ShoutParser! !

!PG3SchemaMirror class methodsFor: 'compiling' stamp: 'klub 3/7/2016 19:14'!
sourceCodeTemplate

	self == PG3SchemaMirror ifTrue: [ ^super sourceCodeTemplate ].
	^'messageSelectorAndArgumentNames

	<pg3Function: ''nameOfTheFunction''
		arguments: #(''argumentName'' #andType)
		returns: #returnType
		volatility: #volatile>
		
begin
	-- here comes the body
end;'
		! !

!PG3SchemaMirror methodsFor: 'as yet unclassified' stamp: 'kb 8/4/2010 23:21'!
executeTransaction: aBlock

	^self class executeTransaction: aBlock! !
